"""Utility functions for aiding plotting."""

import numpy as np
from scipy import stats as sp_stats


class Limiter(object):
    """Helper class to determine limits of data."""

    def __init__(self, limits=None):
        """Initialize the limit gathering."""
        if limits is not None:
            self.min, self.max = limits
        else:
            self.min, self.max = +float('inf'), -float('inf')

    def accumulate(self, data):
        """Analyse data to update limits.

        :param data: new data to analyse.

        """
        self.min = min(self.min, min(data))
        self.max = max(self.max, max(data))
        return self

    def fix(self, lower=None, upper=None):
        """Fix limits to new values.

        :param lower: new lower limit.
        :param upper: new upper limit.

        """
        if lower is not None:
            self.min = lower
        if upper is not None:
            self.max = upper
        return self


def pad(data):
    """Attempt to find a sensible plot bounds for a vector."""
    uniq = sorted(np.unique(data))
    pad = 0.5 * min(np.ediff1d(uniq))
    return uniq[0] - pad, uniq[-1] + pad


def kernel_density_estimate(x, step=0.2):
    """Kernel density to approximate distribution.

    :param x: data of which to find mode.
    :param step: discretization of KDE PDF.
    """
    # estimate bandwidth of kde, R's nrd0 rule-of-thumb
    hi = np.std(x, ddof=1)
    q75, q25 = np.percentile(x, [75, 25])
    iqr = q75 - q25
    lo = min(hi, iqr/1.34)
    if not ((lo == hi) or (lo == abs(x[0])) or (lo == 1)):
        lo = 1
    bw = 0.9 * lo * len(x)**-0.2

    # create a KDE
    x_grid = np.arange(min(x), max(x), step)
    kernel = sp_stats.gaussian_kde(x, bw_method=bw)
    pdf = kernel(x_grid)
    return x_grid, pdf



"""Creation of line plots."""

from aplanat.base import simple


def line(
        x_datas, y_datas, names=None, colors=None,
        xlim=(None, None), ylim=(None, None),
        **kwargs):
    """Create a simple line plot.

    :param x_datas: a list of datasets, each item containing the
        data for that set.
    :params y_datas: a list of weight vectors (one for each dataset).
    :param names: the name of each dataset.
    :param colors: color used to plot each dataset.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    After creating this plot, you may wish to change the following:

        * p.xaxis.axis_label
        * p.yaxis.axis_label
    """
    return simple(
        x_datas, y_datas, names=names, colors=colors,
        xlim=xlim, ylim=ylim, style='line',
        **kwargs)



"""Report building from multiple items."""

from collections import OrderedDict
import uuid

from bokeh.embed import components
from bokeh.model import Model
from bokeh.models.sources import ColumnDataSource
from bokeh.models.widgets import DataTable, TableColumn
from bokeh.resources import INLINE
from jinja2 import Template
import markdown


class HTMLReport(OrderedDict):
    """Generate HTML Report from a series of bokeh figures.

    Items added to the report take an optional key argument, adding items
    with the same key allows an update in place whilst maintaining the order
    in which items were added.
    """

    def __init__(self, title="", lead="", require_keys=False):
        """Initialize the report item collection.

        :param title: report title.
        :param lead: report strapline, shown below title.
        :param require_keys: require keys when adding items.
        """
        self.title = title
        self.lead = lead
        self.require_keys = require_keys
        self.template = Template(
            """\
            <!doctype html>
            <html lang="en">
            <head>
                <meta charset="utf-8">
                <meta http-equiv="X-UA-Compatible" content="IE=edge">
                <meta name="viewport" content="width=device-width, initial-scale=1">
                    <title>{{ title }}</title>
<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/3.4.1/css/bootstrap.min.css" integrity="sha384-HSMxcRTRxnN+Bdg0JdbxYKrThecOKuH5zCYotlSAcp1+c8xmyTe9GYg1l9a69psu" crossorigin="anonymous">

                    {{ resources }}
                    {{ script }}
                </head>
                <body>
            <div class="container">
              <h1>{{ title }}</h1>
              <p class="lead">{{ lead }}
            {{ div }}
                </body>
            </html>
            """  # noqa
        )
        self.plots = list()
        self.md = markdown.Markdown()

    def _add_item(self, item, key=None):
        """Add an item to the report.

        :param item: item to add.
        :param key: unique key for item.
        """
        if key is None:
            if self.require_keys:
                raise ValueError('A key is required.')
            else:
                key = str(uuid.uuid4())
        self[key] = item

    def placeholder(self, key):
        """Add a placeholder to be filled in later."""
        self._add_item(None, key=key)

    def plot(self, plot, key=None):
        """Add a plot to the report.

        :param plot: bokeh plot instance.
        :param key: unique key for item.
        """
        self._add_item(plot, key=key)

    def table(self, df, index=True, key=None, shrink=True, **kwargs):
        """Add a pandas dataframe to the report.

        :param df: pandas dataframe instance.
        :param index: include dataframe index in output.
        :param key: unique key for item.
        :param shrink: shrink wrap the table to avoid whitespace.
        :param kwargs: passed to bokeh DataTable.
        """
        plot = bokeh_table(df, index=index, **kwargs)
        plot.height = min(plot.height, 25*(len(df) + 1))
        self.plot(plot, key)

    def markdown(self, text, key=None):
        """Add markdown formatted text to the report.

        :param text: markdown formatted text.
        :param key: unique key for item.
        """
        html = self.md.convert(text)
        self.md.reset()
        self._add_item(html, key=key)

    def render(self):
        """Generate HTML report containing figures."""
        resources = INLINE.render()
        plots = {k: v for k, v in self.items() if isinstance(v, Model)}
        script, plot_divs = components(plots)

        divs = list()
        for k in self.keys():
            try:
                divs.append(plot_divs[k])
            except KeyError:
                if self[k] is None:
                    raise ValueError(
                        "Placeholder `{}` was not assigned a value.".format(k)
                        )
                divs.append(self[k])
        divs = '\n'.join(divs)
        return self.template.render(
            title=self.title, lead=self.lead,
            resources=resources, script=script, div=divs)

    def write(self, path):
        """Write html report to file."""
        with open(path, "w", encoding='utf8') as outfile:
            outfile.write(self.render())


def bokeh_table(df, index=True, **kwargs):
    """Generate a bokeh table from a pandas dataframe."""
    columns = [TableColumn(field=x, title=x) for x in df.columns]
    if index:
        columns = [TableColumn(field="index", title="Feature")] + columns
    return DataTable(
            columns=columns, source=ColumnDataSource(df), index_position=None,
            **kwargs)



"""Bio specific plots, could be generalised."""

from io import StringIO

from bokeh.models import Range1d
from bokeh.models.tickers import FixedTicker
from bokeh.plotting import figure
import numpy as np
import pandas as pd

# TODO: replace this with data read from a fasta.fai or similar
chrom_data = """
order chrom length
1 1 249250621
2 2 243199373
3 3 198022430
4 4 191154276
5 5 180915260
6 6 171115067
7 7 159138663
8 8 146364022
9 9 141213431
10 10 135534747
11 11 135006516
12 12 133851895
13 13 115169878
14 14 107349540
15 15 102531392
16 16 90354753
17 17 81195210
18 18 78077248
19 19 59128983
20 20 63025520
21 21 48129895
22 22 51304566
23 X 155270560
24 Y 59373566
"""
chrom_data = pd.read_csv(StringIO(chrom_data), delimiter=" ")


def karyotype(x_datas, y_datas, names=None, colors=None, **kwargs):
    """Create a heatmap from three columns.

    :param x: x-axis coordinates (chromosome position)
    :param y: y-axis coordinates (chromosome)
    :param name: title for z-axis.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    if names is None:
        names = [None] * len(x_datas)
    if colors is None:
        colors = [None] * len(x_datas)
    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)
    p = figure(**defaults)
    width = 0.8
    p.rect(
        chrom_data["length"] // 2, chrom_data["order"],
        chrom_data["length"], width,
        fill_color='white', line_color='black')

    plot_order = dict(zip(chrom_data['chrom'], chrom_data['order']))
    for x, y, name, color in zip(x_datas, y_datas, names, colors):
        y = np.array([plot_order[i] for i in y])
        kw = {'alpha': 0.2}
        if name is not None:
            kw['legend_label'] = name
        if color is not None:
            kw['color'] = color
        p.segment(
            x, y - width / 2, x, y + width / 2, **kw)

    # set up the axes
    p.xaxis.formatter.use_scientific = False
    p.yaxis.ticker = FixedTicker()
    p.yaxis.ticker.ticks = list(range(1, 25))
    p.yaxis.major_label_overrides = {23: 'X', 24: 'Y'}
    p.x_range = Range1d(
        start=0, end=max(chrom_data['length']),
        bounds=(0, max(chrom_data['length'])))
    miny, maxy = 0, len(chrom_data) + 1
    p.y_range = Range1d(
        start=miny, end=maxy, bounds=(miny, maxy))
    return p



"""Simple bokeh plotting API."""

import functools
import os
import warnings

import bokeh.io as bkio
from bokeh.layouts import gridplot
from bokeh.model import Model
from bokeh.models import Range1d
from bokeh.models.annotations import Label
from bokeh.plotting import figure
import fontawesome.icons as fontawesome
from pkg_resources import resource_filename
from si_prefix import si_format

__version__ = '0.1.0'

# we don't run a comprehensive test suite and mostly in notebooks,
# so show warnings all the time.
warnings.simplefilter('always', DeprecationWarning)


def with_fontawesome(f):
    """Decorate a plotting function ensuring font awesome is available.

    The function achieves two things: i) triggers the building of bokeh
    custom model which simply includes font awesome and ii) manipulates
    the current working directory to ensure when the extension build is
    triggered (by a call to bokeh.show) we are running somewhere that is
    writeable, otherwise node complains.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        wd = os.getcwd()
        os.chdir(os.path.expanduser("~"))

        class LoadFontAwesome(Model):
            """Interpreting this will make the FontAwesome fonts available."""

            __implementation__ = resource_filename(
                "aplanat", "fontawesome_icon.ts")
            __dependencies__ = {"font-awesome": "^4.6.3"}

        exc = None
        try:
            val = f(*args, **kwargs)
        except Exception as e:
            exc = e
        os.chdir(wd)

        if exc is not None:
            raise exc
        return val
    return wrapper


@with_fontawesome
def bootstrap_fontawesome():
    """Create a 'plot' just to bootstrap fontawesome."""
    p = figure(
        output_backend='webgl',
        plot_width=200, plot_height=10,
        title=None, toolbar_location=None)
    p.axis.visible = False
    p.grid.visible = False
    p.outline_line_color = None
    p.rect([0.5], [0.5], [1.0], [1.0], fill_color=None)
    p.add_layout(
        Label(
            x=0, y=0.3, text_font="FontAwesome", text_color=None,
            text="{} FontAwesome loaded".format(fontawesome['check'])))
    show(p)


class Grid(list):
    """Create a list of lists of fixed length."""

    def __init__(self, width=4):
        """Initialize the Grid."""
        self.width = width
        self.append(list())

    def add(self, item):
        """Add an item to the grid.

        :param item: item to add

        """
        self[-1].append(item)
        if len(self[-1]) == self.width:
            self.append(list())

    def extend(self, items):
        """Add multiple items.

        :param items: items to add.

        """
        for item in items:
            self.add(item)


def show(plot):
    """Show a plot in a notebook."""
    bkio.output_notebook(hide_banner=True)
    bkio.show(plot)


def grid(plots, ncol=4, display=True, **kwargs):
    """Show a grid of plots in a notebook.

    :param plots: a list of bokeh plots.
    :param ncol: number of columns.
    :param kwargs: kwargs for bokeh `gridplot`.

    """
    warnings.warn(
        "Please use `bokeh.layout.gridplot` directly with an optional "
        "call to `.show()`", DeprecationWarning)

    grid = Grid(width=ncol)
    grid.extend(plots)
    plot = gridplot(grid, **kwargs)
    if display:
        show(plot)
    else:
        return plot


class InfoGraphItems(dict):
    """Helper class to cumulatively create items for and infographic."""

    def __init__(self, *args, **kwargs):
        """Initialize the helper, and bootstrap fontawesome."""
        try:
            bootstrap_fontawesome()
        except ImportError:
            pass
        super().__init__(*args, **kwargs)

    def append(self, label, value, icon, unit=''):
        """Add an item.

        :param label: infographic item label.
        :param value: numerical value of headline number (without SI units).
        :param icon: font-awesome icon to use.
        :param unit: additional suffix after SI unit suffix, e.g. "bases".

        """
        self[label] = (label, value, icon, unit)

    def extend(self, items):
        """Add multiple items at once.

        :param items: iterable of 4-tuples, as required by `.add()`.
        """
        for i in items:
            self.append(*i)


@with_fontawesome
def infographic(items, **kwargs):
    """Create and infographic 'plot'.

    :param items: 3-tuples of (label, value, unit, icon); the label should be
        a one or two word description, the value the headline number, and the
        icon the name of a fontawesome icon. `value` should be numeric, it
        will be normalised by use of an SI suffix for display after which
        `unit` will be appended.
    :param kwargs: kwargs for bokeh gridplot.

    ..note:: If `bootstrap_fontawesome` has not already been called, the
        function will load the required fonts, however they will not display
        the first time an Jupyter labs cell is run. If using the
        `InfoGraphItems` helper class, this wrinkle will be taken care of
        provided the helper is initiated in a previous cell.

    """
    plots = list()
    seen = set()
    for label, value, icon, unit in items:
        if label in seen:
            continue
        value = si_format(value) + unit
        seen.add(label)
        p = figure(
            output_backend='webgl',
            plot_width=175, plot_height=100,
            title=None, toolbar_location=None)
        p.axis.visible = False
        p.grid.visible = False
        p.outline_line_color = None
        p.rect([0.5], [0.5], [1.0], [1.0], fill_color="#2171b5")
        p.x_range = Range1d(start=0.1, end=0.9, bounds=(0.1, 0.9))
        p.y_range = Range1d(start=0.1, end=0.9, bounds=(0.1, 0.9))
        p.add_layout(
            Label(
                x=0.15, y=0.45, text=value, text_color="#DEEBF7",
                text_font_size="24px"))
        p.add_layout(
            Label(
                x=0.15, y=0.2, text=label, text_color="#C6DBEF",
                text_font_size="16px"))
        p.add_layout(
            Label(
                x=0.65, y=0.4, text=fontawesome[icon], text_color="#6BAED6",
                text_font="FontAwesome", text_font_size="48px"))
        plots.append(p)
    defaults = {'toolbar_location': None}
    defaults.update(kwargs)
    return gridplot(plots, **defaults)



"""Histogram-like plots."""

from bokeh.models import Range1d
from bokeh.plotting import figure
import numpy as np

from aplanat.util import Limiter


def histogram(
        datas, weights=None, names=None, colors=None,
        normalize=False, bins=30, binwidth=None,
        xlim=(None, None), ylim=(None, None), style='bars',
        **kwargs):
    """Create a histogram plot.

    :param datas: a list of datasets, each item containing the data
        for that set.
    :params weights: a list of weight vectors (one for each dataset).
    :param names: the name of each dataset.
    :param colors: color used to plot each dataset.
    :param normalize: normalize histogram counts by total weight (across all
        datasets).
    :param bins: number of histogram bins.
    :param binwidth: overrides `bins`, provide a precise binwidth.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param style: one of: `bars`, `lines`. `bars` will plot the classical
        bar histogram with a bar depicting counts per fixed interval. `lines`
        adapts this to plot a line per dataset rather than bars.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    After creating this plot, you may wish to change the following:

        * p.xaxis.axis_label
        * p.yaxis.axis_label
    """
    # TODO: slacken this
    if not isinstance(datas, list):
        raise ValueError(
            "datas and weights should be lists, one item per line to plot.")
    items = len(datas)
    if weights is None:
        weights = [None] * items
    if names is None:
        names = [None] * items
    if colors is None:
        colors = ['blue'] * items
    if any(len(x) != items for x in (weights, names, colors)):
        raise IndexError(
            "Lengths of datas, weights, names, and colors should be equal.")

    x_lim = Limiter()
    y_lim = Limiter()
    for data in datas:
        x_lim.accumulate(data)
    if binwidth is not None:
        bins = np.arange(x_lim.min, x_lim.max + 1e-10, binwidth)
    else:
        bins = np.linspace(x_lim.min, x_lim.max, num=bins)
    if normalize:
        total_weight = sum(sum(x) for x in weights)

    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)

    p = figure(**defaults)
    for data, weight, name, color in zip(datas, weights, names, colors):
        counts, edges = np.histogram(data, weights=weight, bins=bins)
        if normalize:
            counts = counts.astype(float) / total_weight
        y_lim.accumulate(counts)
        kw = {}
        if name is not None:
            kw['legend_label'] = name
        if color is not None:
            kw['color'] = color
        if style == 'bars':
            p.quad(
                top=counts, bottom=0, left=edges[:-1], right=edges[1:],
                alpha=0.6, **kw)
        elif style == 'line':
            p.line(x=edges[:-1], y=counts, line_width=1.5, **kw)
    x_lim.fix(*xlim)
    y_lim.fix(*ylim)

    # limit the display range
    p.x_range = Range1d(
        start=x_lim.min, end=x_lim.max, bounds=(x_lim.min, x_lim.max))
    p.y_range = Range1d(
        start=y_lim.min, end=y_lim.max, bounds=(y_lim.min, y_lim.max))
    return p



"""Plotting of spatial/3D data."""

from bokeh.models import ColorBar, LinearColorMapper, Range1d
from bokeh.palettes import Blues9
from bokeh.plotting import figure
from bokeh.transform import linear_cmap, log_cmap
from bokeh.util.hex import hexbin
import numpy as np
import pandas as pd

from aplanat import util


def heatmap(x, y, z, name=None, **kwargs):
    """Create a heatmap from three columns.

    :param x: x-axis coordinates.
    :param y: y-axis coordinates.
    :param z: z-axis coordinates (e.g. counts or frequencies).
    :param name: title for z-axis.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    data = {'x': x, 'y': y, 'z': z}

    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)
    p = figure(**defaults)

    # define a colour map
    colors = Blues9[::-1]
    mapper = LinearColorMapper(
        palette=colors, low=min(data['z']), high=max(data['z']))

    # plot heatmap rectangles
    p.rect(
        source=data, x="x", y="y", width=1, height=1,
        fill_color={'field': 'z', 'transform': mapper},
        line_color=None)

    # add colour scale
    color_bar = ColorBar(
        title=name, color_mapper=mapper, label_standoff=10,
        location=(0, 0))
    p.add_layout(color_bar, 'right')

    # hide some plotting artefacts
    p.x_range.range_padding = 0
    p.y_range.range_padding = 0
    p.xaxis.visible = False
    p.yaxis.visible = False
    # set up the axes
    xbounds = util.pad(data['x'])
    ybounds = util.pad(data['y'])
    p.x_range = Range1d(start=xbounds[0], end=xbounds[1], bounds=xbounds)
    p.y_range = Range1d(start=ybounds[0], end=ybounds[1], bounds=ybounds)
    return p


def heatmap2(
        x, y, name=None, x_bins=50, y_bins=50, log=False,
        xlim=(None, None), ylim=(None, None), **kwargs):
    """Create a heatmap from two columns, using integrate binning.

    In contrast to `heatmap` which takes coordinates and counts, this function
    takes only coordinates and will count observations within bins.

    :param x: x-axis coordinates.
    :param y: y-axis coordinates.
    :param name: title for z-axis.
    :param x_bins: number of bins in x-direction.
    :param y_bins: number of bins in y-direction.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param log: use a log-scaled colour map.

    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    hist, x_edges, y_edges = np.histogram2d(x, y, bins=(x_bins, y_bins))
    table = np.empty(
        x_bins * y_bins,
        dtype=[('row', float), ('col', float), ('value', float)])
    table['value'] = hist.flatten()
    table['row'] = np.repeat(x_edges[:x_bins], y_bins)
    table['col'] = np.tile(y_edges[:y_bins], x_bins)
    table = pd.DataFrame(table)

    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)
    p = figure(**defaults)
    # define a colour map
    cmap = log_cmap if log else linear_cmap
    mapper = cmap('counts', 'Viridis256', 0, max(table['value']))
    # plot heatmap rectangles
    width = (x_edges[-1] - x_edges[0]) / x_bins
    height = (y_edges[-1] - y_edges[0]) / y_bins
    p.rect(
        x="row", y="col", source=table,
        height=height * 1.05, width=width * 1.05,  # fudge to hide artefacts
        fill_color={'field': 'value', 'transform': mapper['transform']},
        line_color=None)
    # add colour scale
    color_bar = ColorBar(
        title=name, color_mapper=mapper['transform'], label_standoff=10,
        location=(0, 0))
    p.add_layout(color_bar, 'right')
    # hide some plotting artefacts
    p.x_range.range_padding = 0
    p.y_range.range_padding = 0
    # set up the axes
    xbounds = util.pad(x_edges[:x_bins])
    ybounds = util.pad(y_edges[:y_bins])
    x_lim = util.Limiter().accumulate(xbounds).fix(*xlim)
    y_lim = util.Limiter().accumulate(ybounds).fix(*ylim)
    p.x_range = Range1d(
        start=x_lim.min, end=x_lim.max, bounds=(x_lim.min, x_lim.max))
    p.y_range = Range1d(
        start=y_lim.min, end=y_lim.max, bounds=(y_lim.min, y_lim.max))
    return p


def hexmap(
        x, y, name=None, log=False, match_aspect=True, kernel_size=0.1,
        xlim=(None, None), ylim=(None, None), **kwargs):
    """Create a heatmap from two columns.

    :param x: x-axis coordinates.
    :param y: y-axis coordinates.
    :param name: title for z-axis.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    """
    bins = hexbin(x, y, kernel_size)

    defaults = {
        "output_backend": "webgl",
        "height": 400, "width": 500,
        "background_fill_color": "#440154"}
    defaults.update(kwargs)
    p = figure(**defaults)

    cmap = log_cmap if log else linear_cmap
    mapper = cmap('counts', 'Viridis256', 0, max(bins.counts))
    p.hex_tile(
        q="q", r="r", size=kernel_size, line_color=None, source=bins,
        fill_color=mapper)

    color_bar = ColorBar(
        title="observations", color_mapper=mapper['transform'],
        label_standoff=10, location=(0, 0))
    p.add_layout(color_bar, 'right')
    p.grid.visible = False

    x_lim = util.Limiter().accumulate(x).fix(*xlim)
    y_lim = util.Limiter().accumulate(y).fix(*ylim)
    # limit the display range
    p.x_range = Range1d(
        start=x_lim.min, end=x_lim.max, bounds=(x_lim.min, x_lim.max))
    p.y_range = Range1d(
        start=y_lim.min, end=y_lim.max, bounds=(y_lim.min, y_lim.max))
    return p, bins



"""Base functions for simple plots."""

from bokeh.models import Range1d
from bokeh.plotting import figure

from aplanat.util import Limiter


def simple(
        x_datas, y_datas, names=None, colors=None,
        xlim=(None, None), ylim=(None, None), style='line',
        **kwargs):
    """Create a simple line or scatter plot.

    :param x_datas: a list of datasets, each item containing the
        data for that set.
    :params y_datas: a list of weight vectors (one for each dataset).
    :param names: the name of each dataset.
    :param colors: color used to plot each dataset.
    :param xlim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param ylim: tuple for plotting limits (start, end). A value None will
        trigger calculation from the data.
    :param style: one of `line` or `points`.
    :param kwargs: kwargs for bokeh figure.

    :returns: a bokeh plot.

    After creating this plot, you may wish to change the following:

        * p.xaxis.axis_label
        * p.yaxis.axis_label
    """
    # TODO: slacken this
    if not isinstance(x_datas, list) or not isinstance(y_datas, list):
        raise ValueError(
            "x_datas and y_datas should be lists, one item per line to plot.")
    items = len(x_datas)
    if names is None:
        names = [None] * items
    if colors is None:
        colors = ['blue'] * items
    if any(len(x) != items for x in (y_datas, names, colors)):
        raise IndexError(
            "Lengths of x_datas, y_datas, names, and colors should be equal.")

    defaults = {
        "output_backend": "webgl",
        "height": 300, "width": 600}
    defaults.update(kwargs)

    p = figure(**defaults)
    x_lim = Limiter()
    y_lim = Limiter()
    for x, y, name, color in zip(x_datas, y_datas, names, colors):
        # TODO: add colours and names
        x_lim.accumulate(x)
        y_lim.accumulate(y)
        kw = {}
        if name is not None:
            kw['legend_label'] = name
        if color is not None:
            kw['color'] = color
        if style == 'line':
            p.line(x=x, y=y, line_width=1.5, **kw)
        elif style == 'points':
            p.circle(x=x, y=y, alpha=0.4, **kw)
        else:
            raise ValueError('Unknown plot style: "{}"'.format(style))
    x_lim.fix(*xlim)
    y_lim.fix(*ylim)

    # limit the display range
    p.x_range = Range1d(
        start=x_lim.min, end=x_lim.max, bounds=(x_lim.min, x_lim.max))
    p.y_range = Range1d(
        start=y_lim.min, end=y_lim.max, bounds=(y_lim.min, y_lim.max))
    return p



"""Add annotations to plots."""

from bokeh.models import Span, Title
import numpy as np


def marker_vline(plot, x, label=None, color='black', width=1.5):
    """Add a vertical line to a plot with optional label.

    :param plot: a bokeh plot instance.
    :param x: the x-axis value where the line should be added.
    :param label: text label.
    :param color: colour for line.
    :param width: line width.

    :returns: updated plot.

    """
    vline = Span(
        location=x, dimension='height',
        line_color=color, line_width=width)
    plot.renderers.extend([vline])
    if label is not None:
        ymax = plot.y_range.end
        if ymax is None:
            raise ValueError(
                "Please set plot.y_range before adding a vline with text.")
        plot.text(
            x=[x], y=[0.5 * ymax], angle=np.pi / 2, text=[label])
    return plot


def marker_hline(plot, y, label=None, color='black', width=1.5):
    """Add a horizontal line to a plot with optional label.

    :param plot: a bokeh plot instance.
    :param y: the y-axis value where the line should be added.
    :param label: text label.
    :param color: colour for line.
    :param width: line width.

    :returns: updated plot.

    """
    hline = Span(
        location=y, dimension='width',
        line_color=color, line_width=width)
    plot.renderers.extend([hline])
    if label is not None:
        xmax = plot.x_range.end
        if xmax is None:
            raise ValueError(
                "Please set plot.x_range before adding a hline with text.")
        plot.text(
            x=[0.5 * xmax], y=[y], angle=0, text=[label])
    return plot


def subtitle(plot, subtitle):
    """Add a subtitle to a plot.

    :param subtitle: the title to add

    :returns: updated plot.
    """
    plot.add_layout(
        Title(text=subtitle, text_font_style='italic'), 'above')
    return plot
